from typing import Type, Union
from mindflow.db.objects.static_definition.model import (
    ModelID,
    ModelTextEmbeddingOpenAI,
    ModelTextCompletionOpenAI,
    ModelOpenAI,
)

from mindflow.utils.enum import ExtendedEnum


class ServiceID(ExtendedEnum):
    OPENAI = "openai"


class ServiceParameterKey(ExtendedEnum):
    ID = "id"
    NAME = "name"
    URL = "url"
    API_URL = "api_url"
    DEFAULT_INDEX_MODEL = "default_index_model"
    DEFAULT_QUERY_MODEL = "default_query_model"
    DEFAULT_EMBEDDING_MODEL = "default_embedding_model"

class ServiceConfigParameterKey(ExtendedEnum):
    API_KEY = "api_key"
    API_SECRET = "api_secret"

    MODELS = "models"


class ServiceConfigParameterName(ExtendedEnum):
    API_KEY = "API Key"
    API_SECRET = "API Secret"

    MODELS = "Models"


class ServiceName(ExtendedEnum):
    OPENAI = "OpenAI"


class ServiceURL(ExtendedEnum):
    OPENAI = ""

class ServiceAPIURL(ExtendedEnum):
    OPENAI = ""

class ServiceDefaultIndexModel(ExtendedEnum):
    OPENAI = ModelID.GPT_3_5_TURBO.value

class ServiceDefaultQueryModel(ExtendedEnum):
    OPENAI = ModelID.GPT_3_5_TURBO.value

class ServiceDefaultEmbeddingModel(ExtendedEnum):
    OPENAI = ModelID.TEXT_EMBEDDING_ADA_002.value

## Service Model Types
class ServiceModelTypeTextEmbedding(ExtendedEnum):
    OPENAI = ModelTextEmbeddingOpenAI


class ServiceModelTypeTextCompletion(ExtendedEnum):
    OPENAI = ModelTextCompletionOpenAI


class ServiceModel(ExtendedEnum):
    OPENAI = ModelOpenAI


SERVICE_STATIC = {
    ServiceID.OPENAI.value: {
        ServiceParameterKey.ID.value: ServiceID.OPENAI.value,
        ServiceParameterKey.NAME.value: ServiceName.OPENAI.value,
        ServiceParameterKey.URL.value: ServiceURL.OPENAI.value,
        ServiceParameterKey.API_URL.value: ServiceURL.OPENAI.value,
        ServiceParameterKey.DEFAULT_INDEX_MODEL.value: ServiceDefaultIndexModel.OPENAI.value,
        ServiceParameterKey.DEFAULT_QUERY_MODEL.value: ServiceDefaultQueryModel.OPENAI.value,
        ServiceParameterKey.DEFAULT_EMBEDDING_MODEL.value: ServiceDefaultEmbeddingModel.OPENAI.value,
    },
}


ServiceUnion = Union[
    ServiceID,
    ServiceParameterKey,
    ServiceConfigParameterKey,
    ServiceName,
    ServiceURL,
    ServiceModel,
    ServiceModelTypeTextEmbedding,
    ServiceModelTypeTextCompletion,
]


def get_service_static(static: Type[ServiceUnion], key: ServiceUnion) -> ServiceUnion:
    return static.__members__[key.name]



from mindflow.utils.enum import ExtendedEnum


class DocumentArgument(ExtendedEnum):
    """
    Document argument enum
    """

    DOCUMENT_TYPE: str = "document_type"
    PATH: str = "path"
    HASH: str = "hash"
    INDEX_TYPE: str = "index_type"
    SEARCH_TREE: str = "search_tree"
    SIZE: str = "size"


class DocumentType(ExtendedEnum):
    """
    Document type enum
    """

    FILE: str = "file"
    SLACK: str = "slack"
    GITHUB: str = "github"
    JIRA: str = "jira"



"""
This file contains the prompts that are used in the CLI.
"""

GIT_DIFF_PROMPT_PREFIX = "The following is a git diff from my code repository. \
    Please thoroughly explain to me in bullet points the changes for each file. \
        Only provide the file name with its bulleted summary and nothing else in the format: \
            {\
                - *file_name*,\
                    - change1 \
                    - change2 \
                    - change3 \
            }\
        with a new line for each file and change."
INDEX_PROMPT_PREFIX = "Pretend you are a search engine trying to provide an information rich \
    yet condensed string that can serve as an index for the contents or the purpose \
    of a file. I want you to respond in as few words as possible while still conveying \
         the full content and purpose of this file."
COMMIT_PROMPT_PREFIX = "Please provide a commit message for the following changes. Only respond with the commit message and nothing else."
PR_TITLE_PREFIX = "Please provide a title for the following pull request using this git diff summary. Only respond with the title and nothing else."
PR_BODY_PREFIX = "Please provide a body for the following pull request using this git diff summary. I'd like it to include, bulleted summaries of \
    changes and titles to give a coherent picture of what has changed. Only respond with the body and nothing else."
QUERY = "Please answer the following query like a helpful virtual assistant using the context provided below: "



from mindflow.db.db.database import Collection
from mindflow.db.objects.base import BaseObject, StaticObject
from mindflow.db.objects.static_definition.service import ServiceID

class Service(StaticObject):
    id: str
    name: str
    url: str
    api_url: str

    _collection: Collection = Collection.SERVICE

class ServiceConfig(BaseObject):
    """Service config object."""
    id: str
    api_key: str
    api_secret: str

    _collection: Collection = Collection.CONFIGURATIONS


class ConfiguredService:
    id: str
    name: str
    url: str
    api_url: str

    api_key: str
    api_secret: str

    def __init__(self, service_id: str):
        service = Service.load(service_id)
        service_config = ServiceConfig.load(f"{service_id}_config")

        if service:
            for key, value in service.__dict__.items():
                setattr(self, key, value)
            
        if service_config:
            for key, value in service_config.__dict__.items():
                if value not in [None, ""]:
                    setattr(self, key, value)
        
        

class ConfiguredServices:
    @property
    def openai(self) -> Service:
        return ConfiguredService(ServiceID.OPENAI.value)



"""
`auth` command
"""
from mindflow.cli.config.main import set_configuration

import click


@click.command(help="Advanced setup custom mindflow configurations.")
def config():
    set_configuration()



"""
`diff` command
"""
import subprocess
from typing import List, Tuple

from mindflow.db.objects.model import Model
from mindflow.settings import Settings
from mindflow.utils.prompt_builders import build_context_prompt

from mindflow.utils.prompts import GIT_DIFF_PROMPT_PREFIX

import concurrent.futures

from mindflow.utils.response import handle_response_text

def run_diff(args: str):
    """
    This function is used to generate a git diff response by feeding git diff to gpt.
    """
    command = ['git', 'diff'] + list(args)

    settings = Settings()
    completion_model: Model = settings.mindflow_models.query.model

    # Execute the git diff command and retrieve the output as a string
    diff_result = subprocess.check_output(command).decode("utf-8")
    batched_parsed_diff_result = batch_git_diffs(parse_git_diff(diff_result), token_limit=completion_model.hard_token_limit)

    response: str = ""
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []
        for batch in batched_parsed_diff_result:
            content = ""
            for (file_name, diff_content) in batch:
                content += f"*{file_name}*\n DIFF CONTENT: {diff_content}\n\n"
            future = executor.submit(completion_model, build_context_prompt(GIT_DIFF_PROMPT_PREFIX, content))
            futures.append(future)

        # Process the results as they become available
        for future in concurrent.futures.as_completed(futures):
            response += future.result()
    
    return response

import re

def parse_git_diff(diff_output: str) -> List[Tuple[str, str]]:
    file_diffs = []
    current_diff = None
    for line in diff_output.split('\n'):
        if line.startswith('diff --git'):
            if current_diff is not None:
                file_diffs.append(current_diff)
            current_diff = {'file_name': None, 'content': []}
            match = re.match(r'^diff --git a/(.+?) b/.+?$', line)
            if match:
                current_diff['file_name'] = match.group(1)
        if current_diff is not None:
            current_diff['content'].append(line)
    if current_diff is not None:
        file_diffs.append(current_diff)
    return [(diff['file_name'], '\n'.join(diff['content'])) for diff in file_diffs]


def batch_git_diffs(file_diffs: List[Tuple[str, str]], token_limit: int) -> List[List[Tuple[str, str]]]:
    batches = []
    current_batch = []
    current_batch_size = 0
    for (file_name, diff_content) in file_diffs:
        if len(diff_content) > token_limit:
            chunks = [diff_content[i:i + token_limit] for i in range(0, len(diff_content), token_limit)]
            for chunk in chunks:
                if current_batch_size + len(chunk) > token_limit * 2:
                    batches.append(current_batch)
                    current_batch = []
                    current_batch_size = 0
                current_batch.append((file_name, chunk))
                current_batch_size += len(chunk)
        elif current_batch_size + len(diff_content) > token_limit * 2:
            batches.append(current_batch)
            current_batch = [(file_name, diff_content)]
            current_batch_size = len(diff_content)
        else:
            current_batch.append((file_name, diff_content))
            current_batch_size += len(diff_content)
    if current_batch:
        batches.append(current_batch)
    return batches



"""
File/Directory Resolver
"""
import os

from typing import List, Optional, Union

from mindflow.resolving.resolvers.base_resolver import BaseResolver
from mindflow.db.objects.document import Document, DocumentReference
from mindflow.db.objects.static_definition.document import DocumentType

from mindflow.utils.files.extract import extract_files


class FileResolver(BaseResolver):
    """
    Resolver for file or directory paths to text.
    """

    @staticmethod
    def should_resolve(document_path: Union[str, os.PathLike]) -> bool:
        """
        Check if a path is a file or directory.
        """
        return os.path.isfile(document_path) or os.path.isdir(document_path)

    def resolve(self, document_path: str) -> List[DocumentReference]:
        """
        Extract text from files.
        """
        document_references: List[DocumentReference] = []
        for path in extract_files(document_path):
            document_reference: Optional[DocumentReference] = None
            document = Document.load(path)

            if document:
                document_reference = document.to_document_reference()
            else:
                document_reference = DocumentReference.from_path(path, DocumentType.FILE)
            
            if document_reference:
                document_references.append(document_reference)
    
        return document_references



from mindflow.cli.config.menu import menu
from mindflow.db.objects.static_definition.mind_flow_model import (
    MindFlowModelID,
    MindFlowModelName,
)
from mindflow.db.objects.static_definition.model import (
    ModelConfigParameterKey,
    ModelConfigParameterName,
    ModelID,
    ModelName,
)
from mindflow.db.objects.static_definition.service import (
    ServiceID,
    ServiceModel,
    get_service_static,
)
from mindflow.db.objects.static_definition.service import (
    ModelTextCompletionOpenAI,
    ModelTextEmbeddingOpenAI,
)


def ask_mind_flow_model_type() -> MindFlowModelID:
    mindflow_type_choice = menu(
        MindFlowModelName.values(), "Which LLM request would you like to configure?"
    )
    return MindFlowModelID[MindFlowModelName(mindflow_type_choice).name]


def ask_model_text_completion_openai() -> ModelID:
    openAIModels = [ModelName[key].value for key in ModelTextCompletionOpenAI.keys()]
    model_choice = menu(openAIModels, "Please select text completion model:")
    return ModelID[ModelName(model_choice).name]


def ask_model_text_embedding_openai() -> ModelID:
    openAIModels = [ModelName[key].value for key in ModelTextEmbeddingOpenAI.keys()]
    model_choice = menu(openAIModels, "Please select text embedding model:")
    return ModelID[ModelName(model_choice).name]


def ask_model_config() -> ModelConfigParameterKey:
    choice = menu(
        ModelConfigParameterName.values(), "Please select a model configuration:"
    )
    return ModelConfigParameterKey[ModelConfigParameterName(choice).name]


def ask_model_by_service(service_key: ServiceID) -> ModelID:
    service_models = get_service_static(ServiceModel, service_key).value
    service_model_names = [ModelName[key].value for key in service_models.keys()]
    model_choice = menu(service_model_names, "Please select a model:")
    return ModelID[ModelName(model_choice).name]



import subprocess

from mindflow.core.diff import run_diff
from mindflow.settings import Settings
from mindflow.utils.prompt_builders import build_context_prompt
from mindflow.utils.prompts import COMMIT_PROMPT_PREFIX

def run_commit() -> str:
    """
    Commit command.
    """
    settings = Settings()

    if not has_staged_files():
        print("No staged files")
        return

    # Execute the git diff command and retrieve the output as a string
    diff_output = run_diff(("--cached",))
    response: str = settings.mindflow_models.query.model(build_context_prompt(COMMIT_PROMPT_PREFIX, diff_output))

    command = ["git", "commit", "-m"] + [response]

    # Execute the git diff command and retrieve the output as a string
    output = subprocess.check_output(command).decode("utf-8")
    return output

def has_staged_files():
    try:
        subprocess.check_call(['git', 'diff', '--cached', '--quiet'])
        return False  # no staged files
    except subprocess.CalledProcessError:
        return True  # there are staged files



from mindflow.db.objects.service import ServiceConfig
from mindflow.db.controller import DATABASE_CONTROLLER

def run_login(openai_api_key):
    service_config = ServiceConfig.load("openai_config")
    if not service_config:
        service_config = ServiceConfig("openai_config")

    service_config.api_key = openai_api_key
    service_config.save()

    DATABASE_CONTROLLER.databases.json.save_file()



import json
import os
import sys

from typing import List, Optional
from mindflow.db.db.database import Database

def _get_mindflow_dir():
    if os.name == 'nt':  # Check if the OS is Windows
        config_dir = os.getenv('APPDATA')
    else:
        config_dir = os.path.join(os.path.expanduser('~'), '.config')
    mindflow_dir = os.path.join(config_dir, 'mindflow')
    return mindflow_dir

MINDFLOW_DIR = _get_mindflow_dir()
if not os.path.exists(MINDFLOW_DIR):
    os.makedirs(MINDFLOW_DIR)

def create_and_load_json(path: str) -> dict:
    if os.path.exists(path):
        # Open the authentication file in read and write mode
        with open(path, "r+", encoding="utf-8") as json_file:
            # Read the existing authentication data
            return json.load(json_file)
    with open(path, "w", encoding="utf-8") as json_file:
        json.dump({}, json_file)
    return {}

JSON_DATABASE_PATH = os.path.join(MINDFLOW_DIR, "db.json")
JSON_DATABASE = create_and_load_json(JSON_DATABASE_PATH)

class JsonDatabase(Database):
    def load(self, collection: str, object_id: str) -> Optional[dict]:
        objects = JSON_DATABASE.get(collection, None)
        if not objects:
            return None

        if isinstance(objects, dict):
            return objects.get(object_id, None)

        return None

    def load_bulk(
        self, collection: str, object_ids: List[str]
    ) -> Optional[List[dict]]:
        objects = JSON_DATABASE.get(collection, None)
        if not objects:
            return None

        if isinstance(objects, dict):
            return [
                objects[object_id] for object_id in object_ids if object_id in objects
            ]

        return None

    def delete(self, collection: str, object_id: str):
        objects = JSON_DATABASE.get(collection, None)
        if not objects:
            return None

        if isinstance(objects, dict):
            if object_id in objects:
                del objects[object_id]

    ### Delete objects from json from ID list and overwrite the file
    def delete_bulk(self, collection: str, object_ids: List[str]):
        if not collection in JSON_DATABASE:
            JSON_DATABASE[collection] = {}

        for object_id in object_ids:
            if JSON_DATABASE[collection].get(object_id, None):
                del JSON_DATABASE[collection][object_id]

    def save(self, collection: str, value: dict):
        if not collection in JSON_DATABASE:
            JSON_DATABASE[collection] = {}

        object_id = value.get("id", None)

        if not object_id:
            print("No ID found in object")
            sys.exit(1)

        JSON_DATABASE[collection][object_id] = value
    
    def save_file(self):
        with open(JSON_DATABASE_PATH, "w", encoding="utf-8") as json_file:
            json.dump(JSON_DATABASE, json_file, indent=4)



from typing import Optional
import openai

from mindflow.db.db.database import Collection
from mindflow.db.objects.base import BaseObject, StaticObject
from mindflow.db.objects.service import ServiceConfig
from mindflow.db.objects.static_definition.model import ModelID
from mindflow.db.objects.static_definition.service import ServiceID

class Model(StaticObject):
    """Model object."""

    id: str
    api: str
    name: str
    service: str
    model_type: str
    hard_token_limit: int
    token_cost: int
    token_cost_unit: str

    # Config
    soft_token_limit: int

    _collection: Collection = Collection.MODEL

class ModelConfig(BaseObject):
    """Model config object."""
    id: str
    soft_token_limit: int

    _collection: Collection = Collection.CONFIGURATIONS

class ConfiguredModel:
    id: str
    name: str
    service: str
    model_type: str
    hard_token_limit: int
    token_cost: int
    token_cost_unit: str

    # Config
    soft_token_limit: int
    api_key: str

    def __init__(self, model_id: str, api_key: str = None):
        model = Model.load(model_id)
        model_config = ModelConfig.load(f"{model_id}_config")
        
        if model:
            for key, value in model.__dict__.items():
                setattr(self, key, value)
            
        if model_config:
            for key, value in model_config.__dict__.items():
                if value not in [None, ""]:
                    setattr(self, key, value)
        
        service_config = ServiceConfig.load(f"{self.service}_config")
        self.api_key = service_config.api_key
    
    def openai_chat_completion(self, messages: list, max_tokens: int = 500, temperature: float = 0.0, stop: Optional[list] = None): 
        openai.api_key = self.api_key
        return openai.ChatCompletion.create(
            model=self.id,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop
        )["choices"][0]["message"]["content"]

    def openai_embedding(self, text: str): 
        openai.api_key = self.api_key
        return openai.Embedding.create(
            engine=self.id,
            input=text
        )["data"][0]["embedding"]

    def __call__(self, prompt, *args, **kwargs):
        if self.service == ServiceID.OPENAI.value:
            if self.id in [ModelID.GPT_3_5_TURBO.value, ModelID.GPT_3_5_TURBO_0301.value]:
                return self.openai_chat_completion(prompt, *args, **kwargs)
            else:
                return self.openai_embedding(prompt, *args, **kwargs)
        else: 
            raise NotImplementedError(f"Service {self.service} not implemented.")



from mindflow.db.objects.mindflow_model import ConfiguredMindFlowModels
from mindflow.db.objects.service import ConfiguredServices


class Settings:
    services: ConfiguredServices
    mindflow_models: ConfiguredMindFlowModels

    def __init__(self):
        self.services = ConfiguredServices()
        self.mindflow_models = ConfiguredMindFlowModels(self.services)



"""
Handles git related operations.
"""
import logging
import os
import subprocess
from typing import List, Union


class NotInGit(BaseException):
    """
    Raised when the given path is not within a git repository.
    """


class GitError(BaseException):
    """
    Raised when a git command fails.
    """


def is_within_git_repo(path: Union[str, os.PathLike]) -> bool:
    """
    Checks if the given path is within a git repository.
    """
    try:
        output = subprocess.run(
            ["git", "-C", path, "rev-parse", "--git-dir"],
            capture_output=True,
            timeout=10,
            check=True,
        )
        return output.returncode == 0
    except NotInGit:
        return False


def get_git_files(path: Union[str, os.PathLike]) -> List[str]:
    """
    Extract all files from a git repository that are not gitignored.
    """
    try:
        output = subprocess.run(
            ["git", "-C", os.fspath(path), "ls-files"],
            capture_output=True,
            timeout=10,
            check=True,
        )
        git_files = output.stdout.decode().strip().split("\n")
        return [os.path.abspath(os.path.join(os.fspath(path), f)) for f in git_files]
    except GitError as error:
        logging.debug("Failed extract git files with 'git ls-files': %s", error)
        return []



"""
Handle response text from Mindflow API.
"""

import pyperclip

def handle_response_text(text: str, skip_clipboard: bool = False):
    """
    Copy to clipboard or print response text.
    """
    # if not skip_clipboard:
    #     try:
    #         pyperclip.copy(f"\n{text}")
    #         print("Response copied to clipboard!!!")
    #         return
    #     except Exception:
    #         # raise Exception(f"Failed to copy to clipboard: {error}")
    #         pass
    print(text)



import sys
from typing import List, Optional

from neo4j import GraphDatabase

from mindflow.db.db.database import Database


class Neo4jDatabase(Database):
    def __init__(self, config: dict):
        self.config = config

    @property
    def driver(self):
        uri = self.config.get("uri", None)
        if not uri:
            print("No uri provided for Neo4j")
            sys.exit(1)
        auth = self.config.get("auth", None)
        if not auth:
            print("No auth provided for Neo4j")
            sys.exit(1)

        user = auth.get("user", None)
        password = auth.get("password", None)
        if not user or not password:
            print("No user or password provided for Neo4j")
            sys.exit(1)

        return GraphDatabase.driver(uri, auth=(user, password))

    @property
    def session(self):
        return self.driver.session()

    def load(self, collection: str, object_id: str) -> Optional[dict]:
        result = self.session.run(
            """
            MATCH (d:{collection}}) WHERE d.id = {object_id}
            RETURN d
        """,
            {"collection": collection, "object_id": object_id},
        )

        return dict(result.single()["d"].items())

    def load_bulk(
        self, collection: str, object_ids: List[str]
    ) -> Optional[List[dict]]:
        result = self.session.run(
            """
            UNWIND {object_ids} AS id
            MATCH (d:{collection}) WHERE d.id = id
            RETURN d
        """,
            {"collection": collection, "object_ids": object_ids},
        )

        return [dict(record["d"].items()) for record in result]

    ### Delete objects from json from ID list and overwrite the file
    def delete_bulk(self, collection: str, object_ids: List[str]):
        return self.session.run(
            """
            UNWIND {object_ids} AS id
            MATCH (d:{collection}) WHERE d.id = id
            DELETE d
        """,
            {"collection": collection, "object_ids": object_ids},
        )

    def save(self, collection: str, value: dict):
        return self.session.run(
            """
            CREATE (d:{collection} {params})
        """,
            {"collection": collection, "params": value},
        )



from typing import List, Union, Optional
from mindflow.db.controller import DATABASE_CONTROLLER
from mindflow.db.db.database import Collection


class StaticObject:
    id: str
    
    _collection: Collection = None

    def __init__(self, id: Union[str, dict]):
        if isinstance(id, dict):
            if not "id" in id:
                raise ValueError("id is required")
            for key, value in id.items():
                setattr(self, key, value)
        else:
            self.id = id
    
    @classmethod
    def load(cls, id: str):
        object_dict: dict = DATABASE_CONTROLLER.databases.static.load(
            cls._collection.value,
            id
        )
        if object_dict is None:
            return None        
        return cls(object_dict)


class BaseObject:
    id: str
    
    _collection: Collection = None

    def __init__(self, id: Union[str, dict]):
        if isinstance(id, dict):
            if not "id" in id:
                raise ValueError("id is required")
            for key, value in id.items():
                setattr(self, key, value)
        else:
            self.id = id

    @classmethod
    def load(self, id: str):
        object_dict: Optional[dict] = DATABASE_CONTROLLER.databases.json.load(
            self._collection.value,
            id
        )
        if object_dict is None:
            return None
        return self(object_dict)
    
    @classmethod
    def load_bulk(cls, ids: list) -> List:
        object_dict: List[dict] = DATABASE_CONTROLLER.databases.json.load_bulk(
            cls._collection.value,
            ids
        )
        if object_dict in [None, []]:
            return []
        return [cls(object) for object in object_dict]
    
    @classmethod
    def delete_bulk(cls, ids: list):
        DATABASE_CONTROLLER.databases.json.delete_bulk(
            cls._collection.value,
            ids
        )
    
    def delete(self):
        DATABASE_CONTROLLER.databases.json.delete(
            self._collection.value,
            self.id
        )
    
    def save(self): 
        DATABASE_CONTROLLER.databases.json.save(self._collection.value, self.todict(self))
    
    def todict(self, obj, classkey=None):
        if isinstance(obj, dict):
            data = {}
            for (k, v) in obj.items():
                data[k] = self.todict(v, classkey)
            return data
        elif hasattr(obj, "_ast"):
            return self.todict(obj._ast(), classkey)
        elif hasattr(obj, "__iter__") and not isinstance(obj, str):
            return [self.todict(v, classkey) for v in obj]
        elif hasattr(obj, "__dict__"):
            data = dict(
                [
                    (key, self.todict(value, classkey))
                    for key, value in obj.__dict__.items()
                    if not callable(value) and not key.startswith("_")
                ]
            )
            if classkey is not None and hasattr(obj, "__class__"):
                data[classkey] = obj.__class__.__name__
            return data
        else:
            return obj



"""
`generate` command
"""

from typing import List
import click

from mindflow.core.index import run_index

@click.command(help="Index path(s). You can pass as many folders/files/paths as you'd like. Pass `.` to reference all ")
@click.argument("document_paths", type=str, nargs=-1, required=True)
@click.option("--force", is_flag=True, default=False)
@click.option("--refresh", is_flag=True, default=False)
def index(document_paths: List[str], refresh: bool, force: bool) -> None:
    run_index(document_paths, refresh, force)


from mindflow.db.objects.static_definition.service import (
    ServiceConfigParameterKey,
    ServiceConfigParameterName,
    ServiceID,
    ServiceName,
)
from mindflow.cli.config.main import menu


def ask_service_authorization() -> ServiceID:
    choice = menu(ServiceName.values(), "Which service would you like to authorize?")
    return ServiceID[ServiceName(choice).name]


def ask_service() -> ServiceID:
    choice = menu(ServiceName.values(), "Please select a service:")
    return ServiceID[ServiceName(choice).name]


def ask_service_config() -> ServiceConfigParameterKey:
    choice = menu(
        ServiceConfigParameterName.values(), "Please select a service configuration:"
    )
    return ServiceConfigParameterKey[ServiceConfigParameterName(choice).name]




import subprocess
from typing import List
from mindflow.core.diff import run_diff
from mindflow.settings import Settings
from mindflow.utils.prompt_builders import build_context_prompt
from mindflow.utils.prompts import PR_BODY_PREFIX, PR_TITLE_PREFIX


def run_pr():

    if not has_remote_branch():
        print("No remote branch for current branch")
        return
    
    if needs_push():
        print("Current branch needs to be pushed to remote repository")
        return

    settings = Settings()

    # Determine the name of the default branch
    default_branch = subprocess.check_output(['git', 'symbolic-ref', 'refs/remotes/origin/HEAD']).decode().strip().split('/')[-1]
    diff_output = run_diff((default_branch,))

    pr_title_prompt = build_context_prompt(PR_TITLE_PREFIX, diff_output)
    pr_title = settings.mindflow_models.query.model(pr_title_prompt)

    pr_body_prompt = build_context_prompt(PR_BODY_PREFIX, diff_output)
    pr_body = settings.mindflow_models.query.model(pr_body_prompt)

    create_pull_request(pr_title, pr_body)

def needs_push() -> bool:
    """
    Returns True if the current branch needs to be pushed to a remote repository, False otherwise.
    """
    # Get the output of `git status`
    git_status = subprocess.check_output(["git", "status"]).decode("utf-8")

    # Check if the output contains the message "Your branch is ahead of 'origin/<branch>' by <num> commit(s), and can be fast-forwarded."
    return "Your branch is ahead of" in git_status

def has_remote_branch() -> bool:
    """
    Returns True if there is a remote branch for the current branch, False otherwise.
    """
    # Get the name of the current branch
    current_branch = subprocess.check_output(["git", "symbolic-ref", "--short", "HEAD"]).decode("utf-8").strip()

    # Check if there is a remote branch for the current branch
    try:
        subprocess.check_output(["git", "ls-remote", "--exit-code", "--heads", "origin", current_branch])
        return True
    except subprocess.CalledProcessError:
        return False

def create_pull_request(title, body):
    command: List[str] = ["gh", "pr", "create", "--title", title, "--body", body]
    pr_result = subprocess.check_output(command).decode("utf-8")
    if "https://" in pr_result:
        print("Pull request created successfully")
        print(pr_result)
    else:
        print("Failed to create pull request. Please raise an issue at: https://github.com/nollied/mindflow-cli/issues")
        exit()



from typing import Type, Union
from mindflow.db.objects.service import Service, ServiceConfig
from mindflow.db.objects.model import Model, ModelConfig
from mindflow.db.objects.mindflow_model import MindFlowModel, MindFlowModelConfig
from mindflow.db.objects.document import Document
from mindflow.utils.enum import ExtendedEnum

class ObjectID(ExtendedEnum): 
    SERVICE = "service"
    SERVICE_CONFIG = "service_config"
    MODEL = "model"
    MODEL_CONFIG = "model_config"
    MINDFLOW_MODEL_CONFIG = "mindflow_model_config"
    DOCUMENT = "document"

class ObjectName(ExtendedEnum):
    SERVICE = "Service"
    SERVICE_CONFIG = "Service Configurations"
    MODEL = "Model"
    MODEL_CONFIG = "Model Configurations"
    MINDFLOW_MODEL_CONFIG = "Mindflow Model Configurations"
    DOCUMENT = "Document"

class ObjectType(ExtendedEnum):
    SERVICE = Service
    SERVICE_CONFIG = ServiceConfig
    MODEL = Model
    MODEL_CONFIG = ModelConfig
    MINDFLOW_MODEL = MindFlowModel
    MINDFLOW_MODEL_CONFIG = MindFlowModelConfig
    DOCUMENT = Document

OBJECT_UNION = Union[Service, ServiceConfig, Model, ModelConfig, MindFlowModelConfig, Document]

def get_object_static(static: Type[OBJECT_UNION], key: OBJECT_UNION) -> OBJECT_UNION:
    return static.__members__[key.name]


import click
from mindflow.core.login import run_login

@click.command(help="Set your OpenAI API Key. You can get this from `https://platform.openai.com/account/api-keys`")
@click.argument("openai_api_key")
def login(openai_api_key):
    run_login(openai_api_key)



"""
`inspect` command
"""
import json
from typing import List
from mindflow.db.controller import DATABASE_CONTROLLER
from mindflow.db.db.database import Collection

from mindflow.resolving.resolve import resolve_all

def run_inspect(document_paths: List[str]):
    """
    This function is used to inspect your MindFlow index.
    """
    document_paths = [document.path for document in resolve_all(document_paths)]
    print(
        json.dumps(
            DATABASE_CONTROLLER.databases.json.load_bulk(
                Collection.DOCUMENT.value, document_paths
            ),
            indent=4,
        )
    )



"""
Module for resolving documents to text.
"""

import sys
from typing import List, Optional

from mindflow.db.objects.document import DocumentReference
from mindflow.input import Command
from mindflow.resolving.resolvers.file_resolver import FileResolver


def resolve(document_path: str) -> List[DocumentReference]:
    """
    Resolves a document to text.
    """

    resolvers = [FileResolver()]
    for resolver in resolvers:
        if resolver.should_resolve(document_path):
            return resolver.resolve(document_path)

    print(f"Cannot resolve document: {document_path}")
    sys.exit(1)

def resolve_all(document_paths: List[str]) -> List[DocumentReference]:
    document_references: List[DocumentReference] = []
    for document_path in document_paths:
        document_references.extend(resolve(document_path))
    return document_references

def return_if_indexable(document_references: List[DocumentReference], refresh: bool, force: bool) -> List[DocumentReference]:
    return [
        document_reference
        for document_reference in document_references
        if index_document(document_reference, refresh, force)
    ]

def index_document(
    document_reference: DocumentReference, refresh: bool, force: Optional[bool]
) -> bool:
    if refresh:
        if not document_reference.old_hash:
            return False
        if document_reference.old_hash == document_reference.hash and not force:
            return False
        return True
  
    if not hasattr(document_reference, "old_hash") or force:
        return True
    return False



"""
Base Resolver Class
"""

from typing import List, Optional

from mindflow.db.objects.document import DocumentReference


class BaseResolver:
    """
    Base class for resolvers
    """

    @staticmethod
    def read_document(document_path: str) -> Optional[str]:
        """
        Read a document.
        """

    @staticmethod
    def should_resolve(document_path: str) -> bool:
        """
        Checks if a string is a valid document path for this resolver.
        """
        return False

    def resolve(self, document_path: str) -> List[DocumentReference]:
        """
        Resolve a document path to text.
        """
        return []



import hashlib
from typing import Optional
from mindflow.db.db.database import Collection
from mindflow.db.objects.base import BaseObject
from mindflow.db.objects.static_definition.document import DocumentType


class Document(BaseObject):
    """
    Document
    """

    id: str
    path: str
    document_type: str
    size: int
    hash: str
    search_tree: dict

    _collection = Collection.DOCUMENT

    def to_document_reference(
        self,
    ) -> "DocumentReference":
        """
        Create document reference
        """
        document_text: Optional[str] = read_document(self.id, self.document_type)
        if not document_text:
            raise Exception("Document text not found")
        
        document_text_bytes = document_text.encode("utf-8")
        return DocumentReference(
            {
                "id": self.id,
                "path": self.path,
                "document_type": self.document_type,
                "size": len(document_text_bytes),
                "hash": hashlib.sha256(document_text_bytes).hexdigest(),
                "old_hash": self.hash,
            }
        )


class DocumentReference(BaseObject):
    """
    Used to handle referenced to indexed documents or yet uncreated documents.
    """

    id: str
    path: str
    document_type: str
    size: int
    hash: str
    old_hash: Optional[str]

    def to_document(
        self,
    ) -> Document:
        """
        Create document
        """
        return Document({
            "id": self.id,
            "path": self.path,
            "document_type": self.document_type,
            "size": self.size,
            "hash": self.hash,
        })

    @classmethod
    def from_path(
        cls,
        document_path: str,
        document_type: DocumentType,
    ) -> Optional["DocumentReference"]:
        """
        Create document reference from path
        """
        document_text: Optional[str] = read_document(document_path, document_type.value)
        if not document_text:
            return None
        document_text_bytes = document_text.encode()
        return cls({
            "id": document_path,
            "path": document_path,
            "document_type": document_type.value,
            "size": len(document_text_bytes),
            "hash": hashlib.sha256(document_text_bytes).hexdigest(),
        })

def read_document(document_path: str, document_type: str) -> Optional[str]:
    """
    Read a document.
    """
    match document_type:
        case DocumentType.FILE.value:
            try:
                with open(document_path, "r", encoding="utf-8") as file:
                    return file.read()
            except UnicodeDecodeError:
                return None
        case _:
            return None


"""
This file is used to load the chat gpt model.
"""

from typing import Optional
import numpy as np
import openai
from mindflow.db.objects.model import Model
from mindflow.db.objects.static_definition.model import ModelID

from mindflow.utils.prompts import SEARCH_INDEX


class GPTEndpoints:
    """
    Class for interacting with OpenAI GPT models
    """

    @property
    def configured_api(self):
        openai.api_key = STATE.settings.services.openai.api_key
        return openai

    def summarize(self, text: str) -> str:
        """
        Get response from OpenAI API
        """
        try:
            # print(f"Suffix: {suffix}")
            if STATE.settings.mindflow_models.index.model.api in [ModelID.GPT_3_5_TURBO.value, ModelID.GPT_3_5_TURBO_0301.value]:
                response = self.configured_api.ChatCompletion.create(
                    model=STATE.settings.mindflow_models.index.model.api,
                    messages=[
                        {"role": "system", "content": SEARCH_INDEX},
                        {"role": "user", "content": text},
                    ],
                    temperature=0,
                    max_tokens=500,
                    stop=["\n\n"],
                )["choices"][0]["message"]["content"]
            else:
                response = self.configured_api.Completion.create(
                    engine=STATE.settings.mindflow_models.index.model.api,
                    prompt=f"{SEARCH_INDEX}\n\n{text}",
                    temperature=0,
                    max_tokens=500,
                )["choices"][0]["text"]

            return response
        except Exception as e:
            print(e)
            return ""

    def query(self, prompt: str, selected_content: Optional[str] = None) -> str:
        """
        Get response from OpenAI API
        """
        try:
            # print(f"Suffix: {suffix}")

            if STATE.settings.mindflow_models.query.model.api in ["gpt-3.5-turbo", "gpt-3.5-turbo-0301"]:
                response = self.configured_api.ChatCompletion.create(
                    model=STATE.settings.mindflow_models.index.model.api,
                    messages=[
                        {"role": "system", "content": "You are a helpful virtual assistant responding to a users query using your general knowledge and the text provided below."},
                        {"role": "user", "content": prompt},
                        {"role": "system", "content": selected_content}
                    ],
                    temperature=0,
                    max_tokens=1000,
                )["choices"][0]["message"]["content"]

                return response
            else:
                response = self.configured_api.Completion.create(
                    engine=STATE.settings.mindflow_models.query.model.api,
                    prompt=prompt,
                    suffix=selected_content,
                    temperature=0,
                    max_tokens=1000,
                )["choices"][0]["text"]
            return response

        except Exception as e:
            print(e)
            return ""

    def embed(self, text: str) -> np.ndarray:
        """
        Get embedding from OpenAI API
        """
        try:
            response = self.configured_api.Embedding.create(
                engine=STATE.settings.mindflow_models.embedding.model.api, input=text
            )["data"][0]["embedding"]

            return response
        except Exception as e:
            print(e)
            return np.array([])


def get_response(configured_api: openai, model: Model, prompt: str, selected_content: Optional[str] = None) -> str:
    response = configured_api.ChatCompletion.create(
                model=model.api,
                messages=[
                    {"role": "system", "content": "You are a helpful virtual assistant responding to a users query using your general knowledge and the text provided below."},
                    {"role": "user", "content": prompt},
                    {"role": "system", "content": selected_content}
                ],
                temperature=0,
                max_tokens=1000,
            )["choices"][0]["message"]["content"]

openai.Model.retrieve()

GPT = GPTEndpoints()



from typing import Optional
from mindflow.db.objects.static_definition.document import DocumentType
from mindflow.utils.files.utf8 import is_valid_utf8


def read_document(doc_path: str, doc_type: str) -> Optional[str]:
    """
    Read document
    """
    if doc_type == DocumentType.FILE.value:
        if is_valid_utf8(doc_path):
            with open(doc_path, "r", encoding="utf-8") as file:
                return file.read()
        return None
    else:
        raise Exception(f"Document type {doc_type} not supported")



from typing import Type, Union
from mindflow.db.objects.static_definition.model_type import ModelType
from mindflow.db.objects.static_definition.model import ModelID
from mindflow.utils.enum import ExtendedEnum

class MindFlowModelParameterKey(ExtendedEnum):
    ID = "id"
    NAME = "name"
    DEFAULTS = "defaults"

class MindFlowModelID(ExtendedEnum):
    QUERY = "query"
    INDEX = "index"
    EMBEDDING = "embedding"

class MindFlowModelDefaults(ExtendedEnum):
    QUERY: dict = {
        "openai": ModelID.GPT_3_5_TURBO.value,
    }
    INDEX: dict = {
        "openai": ModelID.GPT_3_5_TURBO.value,
    }
    EMBEDDING: dict = {
        "openai": ModelID.TEXT_EMBEDDING_ADA_002.value,
    }

class MindFlowModelName(ExtendedEnum):
    QUERY = "Query"
    INDEX = "Index"
    EMBEDDING = "Embedding"

class MindFlowModelType(ExtendedEnum):
    QUERY = ModelType.TEXT_COMPLETION.value
    INDEX = ModelType.TEXT_COMPLETION.value
    EMBEDDING = ModelType.TEXT_EMBEDDING.value

class MindFlowModelDescription(ExtendedEnum):
    QUERY = f"{MindFlowModelName.QUERY}:    Used to generate final response from query."
    INDEX = f"{MindFlowModelName.INDEX}:    Used to generate deep index from query. (Warning! Expensive and slow!)"


MINDFLOW_MODEL_STATIC = {
    MindFlowModelID.QUERY.value: {
        MindFlowModelParameterKey.ID.value: MindFlowModelID.QUERY.value,
        MindFlowModelParameterKey.NAME.value: MindFlowModelName.QUERY.value,
        MindFlowModelParameterKey.DEFAULTS.value: MindFlowModelDefaults.QUERY.value,
    },
    MindFlowModelID.INDEX.value: {
        MindFlowModelParameterKey.ID.value: MindFlowModelID.INDEX.value,
        MindFlowModelParameterKey.NAME.value: MindFlowModelName.INDEX.value,
        MindFlowModelParameterKey.DEFAULTS.value: MindFlowModelDefaults.INDEX.value,
    },
    MindFlowModelID.EMBEDDING.value: {
        MindFlowModelParameterKey.ID.value: MindFlowModelID.EMBEDDING.value,
        MindFlowModelParameterKey.NAME.value: MindFlowModelName.EMBEDDING.value,
        MindFlowModelParameterKey.DEFAULTS.value: MindFlowModelDefaults.EMBEDDING.value,
    }
}

MindFlowModelUnion = Union[
    MindFlowModelID,
    MindFlowModelDefaults,
    MindFlowModelName,
    MindFlowModelType,
    MindFlowModelDescription,
]


def get_mind_flow_model_static(
    static: Type[MindFlowModelUnion], key: MindFlowModelUnion
) -> MindFlowModelUnion:
    return static.__members__[key.name]



import sys
from typing import Dict, Optional
from mindflow.db.objects.base import BaseObject, StaticObject
from mindflow.db.db.database import Collection
from mindflow.db.objects.model import Model, ConfiguredModel
from mindflow.db.objects.service import ConfiguredServices
from mindflow.db.objects.static_definition.mind_flow_model import MindFlowModelID

class MindFlowModel(StaticObject):
    """MindFlow model object."""
    id: str # index, query, embedding
    name: str
    defaults: Dict[str, str]
    model: str

    _collection: Collection = Collection.MIND_FLOW_MODEL
        
class MindFlowModelConfig(BaseObject):
    """MindFlow model config object."""
    id: str
    model: str

    _collection: Collection = Collection.CONFIGURATIONS

class ConfiguredMindFlowModel:
    """MindFlow model object."""
    id: str # index, query, embedding
    name: str
    defaults: Dict[str, str]
    model: Model

    def __init__(self, mindflow_model_id: str, configured_services: ConfiguredServices): 
        self.id = mindflow_model_id

        mind_flow_model = MindFlowModel.load(mindflow_model_id)
        mind_flow_model_config = MindFlowModelConfig.load(f"{mindflow_model_id}_config")

        if mind_flow_model:
            for key, value in mind_flow_model.__dict__.items():
                setattr(self, key, value)

        model_id: Optional[str] = None   
        if mind_flow_model_config:
            if hasattr(mind_flow_model_config, "model"):
                model_id = mind_flow_model_config.model

        if model_id is None:
            model_id = self.get_default_model_id(mindflow_model_id, configured_services)  

        self.model = ConfiguredModel(model_id)
    
    def get_default_model_id(self, mindflow_model_id: str, configured_services: ConfiguredServices) -> str:
        if hasattr(configured_services.openai, "api_key"):
            service = configured_services.openai
        else:
            print("No service API key configured. Please configure an API key for at least one service.")
            sys.exit(1)
        
        model_id = self.defaults.get("openai", None)

        if model_id is None:
            raise Exception("No default model configured for mindflow model: " + mindflow_model_id + " and service: " + service.id)

        return model_id

class ConfiguredMindFlowModels:
    def __init__(self, configured_services: ConfiguredServices):
        self.configured_services = configured_services
    
    @property
    def index(self):
        model = ConfiguredMindFlowModel(MindFlowModelID.INDEX.value, self.configured_services)
        return model

    @property
    def query(self):
        model = ConfiguredMindFlowModel(MindFlowModelID.QUERY.value, self.configured_services)
        return model

    @property
    def embedding(self):
        model = ConfiguredMindFlowModel(MindFlowModelID.EMBEDDING.value, self.configured_services)
        return model



import click

from mindflow.core.commit import run_commit

@click.command(help="Generate a git commit response by feeding git diff to gpt")
def commit():
    """
    Commit command.
    """
    print(run_commit())



"""
`delete` command
"""

from typing import List
from mindflow.db.controller import DATABASE_CONTROLLER
from mindflow.db.objects.document import Document
from mindflow.resolving.resolve import resolve_all

def run_delete(document_paths: List[str]):
    """
    This function is used to delete your MindFlow index.
    """
    Document.delete_bulk([
        document_reference.path for document_reference in resolve_all(document_paths)
    ])
    DATABASE_CONTROLLER.databases.json.save_file()


from flask import Flask, request
from mindflow.db.objects.configurations import Configurations

from mindflow.input import Arguments, Command
from mindflow.settings import Settings
from mindflow.state import STATE

from mindflow.core.ask import ask

# from mindflow.commands.config import config
from mindflow.cli.new_click_cli.commands.delete import delete
from mindflow.cli.new_click_cli.commands.diff import diff
from mindflow.cli.new_click_cli.commands.index import index
from mindflow.cli.new_click_cli.commands.inspect import inspect
from mindflow.core.query import query


def trim_json(data: dict, keys: list) -> dict:
    return {key: data[key] for key in keys if key in data}


class API:
    def __init__(self):
        self.app = Flask(__name__)

        @self.app.route("/ask", methods=["POST"])
        def ask_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["query", "return_prompt", "skip_clipboard"]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            ask()

        @self.app.route("/config", methods=["POST"])
        def config_route():
            # Your implementation for Command.CONFIG
            pass

        @self.app.route("/delete", methods=["POST"])
        def delete_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["document_paths"]

            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            delete()

        @self.app.route("/diff", methods=["POST"])
        def diff_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["git_diff_args", "return_prompt", "skip_clipboard"]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            diff()

        @self.app.route("/inspect", methods=["POST"])
        def inspect_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["document_paths"]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            inspect()

        @self.app.route("/query", methods=["POST"])
        def query_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = [
                "document_paths",
                "index",
                "query",
                "return_prompt",
                "skip_clipboard",
            ]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            query()

        @self.app.route("/refresh", methods=["POST"])
        def refresh_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["document_paths", "force"]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            index()

        @self.app.route("/index", methods=["POST"])
        def index_route():
            params = request.get_json()
            arguments = params.get("arguments", {})
            keys_to_keep = ["document_paths"]
            arguments = trim_json(arguments, keys_to_keep)

            database = params.get("database", None)
            path = params.get("path", None)
            auth = params.get("auth", None)
            user_configurations = params.get("user_configurations", {})

            STATE.user_configurations = Configurations(
                user_configurations
            )
            STATE.settings = Settings.initialize(user_configurations)
            STATE.arguments = Arguments(arguments)
            STATE.command = Command.INDEX.value

            index()


api = API()
app = api.app
app.run()



"""
`query` command
"""

from concurrent.futures import ThreadPoolExecutor, as_completed
from sklearn.metrics.pairwise import cosine_similarity
from typing import Dict, List, Optional, Tuple

import sys
import numpy as np

from mindflow.db.objects.document import Document, DocumentReference
from mindflow.db.objects.model import Model
from mindflow.resolving.resolve import resolve_all
from mindflow.settings import Settings

from mindflow.utils.response import handle_response_text

def run_query(document_paths: List[str], query: str):
    """
    This function is used to ask a custom question about files, folders, and websites.
    """
    settings = Settings()
    completion_model = settings.mindflow_models.query.model
    embedding_model = settings.mindflow_models.embedding.model

    document_references: List[DocumentReference] = resolve_all(document_paths)
    messages = build_query_messages(query, select_content(query, document_references, completion_model.hard_token_limit, embedding_model))
    response = completion_model(messages)
    return response

def build_query_messages(query: str, content: str) -> List[Dict]:
    """
    This function is used to build the query messages for the prompt.
    """
    return [
            {"role": "system", "content": "You are a helpful virtual assistant responding to a users query using your general knowledge and the text provided below."},
            {"role": "user", "content": query},
            {"role": "system", "content": content}
        ]

def select_content(query: str, document_references: List[DocumentReference], token_limit: int, embedding_model: Model) -> str:
    """
    This function is used to generate a prompt based on a question or summarization task
    """
    embedding_ranked_document_chunks: List[
        Tuple(DocumentChunk, float)
    ] = rank_document_chunks_by_embedding(query, document_references, embedding_model)
    if len(embedding_ranked_document_chunks) == 0:
        print(
            "No index for requested hashes. Please generate index for passed content."
        )
        sys.exit(1)

    selected_content = trim_content(embedding_ranked_document_chunks, token_limit)

    return selected_content


class DocumentChunk:
    """
    This class is used to store the chunks of a document.
    """

    def __init__(
        self, path: str, start: int, end: int, embedding: Optional[np.ndarray] = None
    ):
        self.path = path
        self.start = start
        self.end = end
        self.embedding = embedding

    @classmethod
    def from_search_tree(
        cls,
        document: Document,
        embedding_model: Model,
    ) -> Tuple[List["DocumentChunk"], List[np.ndarray]]:
        """
        This function is used to split the document into chunks.
        """

        stack = [document.search_tree]
        chunks: List["DocumentChunk"] = [
            cls(
                document.path,
                document.search_tree["start"],
                document.search_tree["end"],
            )
        ]
        embeddings: List[np.ndarray] = [embedding_model(document.search_tree["summary"])]
        rolling_summary: List[str] = []
        while stack:
            node = stack.pop()
            rolling_summary.append(node["summary"])
            if node["leaves"]:
                for leaf in node["leaves"]:
                    stack.append(leaf)
                    chunks.append(cls(document.path, leaf["start"], leaf["end"]))
                    rolling_summary_embedding = embedding_model(
                        "\n\n".join(rolling_summary) + "\n\n" + leaf["summary"],
                    )
                    embeddings.append(rolling_summary_embedding)
            rolling_summary.pop()

        return chunks, embeddings


def trim_content(ranked_document_chunks: List[DocumentChunk], token_limit: int) -> str:
    """
    This function is used to select the most relevant content for the prompt.
    """
    selected_content: str = ""
    char_limit: int = token_limit * 3

    for document_chunk in ranked_document_chunks:
        if document_chunk:
            with open(document_chunk.path, "r", encoding="utf-8") as file:
                file.seek(document_chunk.start)
                text = file.read(document_chunk.end - document_chunk.start)
                if len(selected_content + text) > char_limit:
                    selected_content += text[: char_limit - len(selected_content)]
                    break
                selected_content += text

    return selected_content


def rank_document_chunks_by_embedding(query: str, document_references: List[DocumentReference], embedding_model: Model) -> List[DocumentChunk]:
    """
    This function is used to select the most relevant content for the prompt.
    """
    prompt_embeddings = np.array(embedding_model(query)).reshape(1, -1)

    ranked_document_chunks = []
    for i in range(0, len(document_references), 100):
        document_ids = [
            document.id for document in document_references[i : i + 100]
        ]
        documents: List[Optional[Document]] = Document.load_bulk(document_ids)
        documents = [document for document in documents if document]
        if documents == []:
            continue
        
        with ThreadPoolExecutor(max_workers=50) as executor:
            futures = [
                executor.submit(DocumentChunk.from_search_tree, document, embedding_model)
                for document in documents 
            ]
            for future in as_completed(futures):
                document_chunks, document_chunk_embeddings = future.result()
                similarities = cosine_similarity(
                    prompt_embeddings, document_chunk_embeddings
                )[0]
                ranked_document_chunks.extend(list(zip(document_chunks, similarities)))

    ranked_document_chunks.sort(key=lambda x: x[1], reverse=True)
    return [chunk for chunk, _ in ranked_document_chunks]



"""
Quick UTF-8 validation.
"""


def is_valid_utf8(document_text: str) -> bool:
    """
    Check if a file is valid utf8.
    """
    try:
        document_text.encode("utf-8")
        return True
    except UnicodeDecodeError:
        return False



from mindflow.utils.enum import ExtendedEnum


class MainOption(ExtendedEnum):
    AUTHORIZE = "authorize a service"
    MIND_FLOW_MODEL = "select model for use by MindFlow"
    MODEL = "configure a model"


class YesNo(ExtendedEnum):
    YES = "Yes"
    NO = "No"



from enum import Enum


class ExtendedEnum(Enum):
    @classmethod
    def keys(cls):
        return list(cls.__members__.keys())

    @classmethod
    def values(cls):
        return list(map(lambda c: c.value, cls))



from typing import Type, Union
from mindflow.db.objects.static_definition.model_type import ModelType
from mindflow.utils.enum import ExtendedEnum


class ModelID(ExtendedEnum):
    GPT_3_5_TURBO = "gpt-3.5-turbo"
    GPT_3_5_TURBO_0301 = "gpt-3.5-turbo-0301"

    # TEXT_DAVINCI_003 = "text-davinci-003"
    # TEXT_CURIE_001 = "text-curie-001"
    # TEXT_BABBAGE_001 = "text-babbage-001"
    # TEXT_ADA_001 = "text-ada-001"

    TEXT_EMBEDDING_ADA_002 = "text-embedding-ada-002"

class ModelParameterKey(ExtendedEnum):
    ID = "id"
    NAME = "name"
    SERVICE = "service"
    MODEL_TYPE = "model_type"
    SOFT_TOKEN_LIMIT = "soft_token_limit"
    HARD_TOKEN_LIMIT = "hard_token_limit"
    TOKEN_COST = "token_cost"
    TOKEN_COST_UNIT = "token_cost_unit"
    DESCRIPTION = "description"

class ModelService(ExtendedEnum):
    GPT_3_5_TURBO = "openai"
    GPT_3_5_TURBO_0301 = "openai"

    # TEXT_DAVINCI_003 = "openai"
    # TEXT_CURIE_001 = "openai"
    # TEXT_BABBAGE_001 = "openai"
    # TEXT_ADA_001 = "openai"

    TEXT_EMBEDDING_ADA_002 = "openai"


class ModelConfigParameterKey(ExtendedEnum):
    SOFT_TOKEN_LIMIT = "soft_token_limit"


class ModelConfigParameterName(ExtendedEnum):
    SOFT_TOKEN_LIMIT = "Soft Token Limit"


class ModelName(ExtendedEnum):
    GPT_3_5_TURBO = "GPT 3.5 Turbo"
    GPT_3_5_TURBO_0301 = "GPT 3.5 Turbo March 1st"

    # TEXT_DAVINCI_003 = "Text Davinci 003"
    # TEXT_CURIE_001 = "Text Curie 001"
    # TEXT_BABBAGE_001 = "Text Babbage 001"
    # TEXT_ADA_001 = "Text Ada 001"

    TEXT_EMBEDDING_ADA_002 = "Text Embedding Ada 002"

class ModelSoftTokenLimit(ExtendedEnum):
    GPT_3_5_TURBO = 500
    GPT_3_5_TURBO_0301 = 500

    # TEXT_DAVINCI_003 = 800
    # TEXT_CURIE_001 = 500
    # TEXT_BABBAGE_001 = 500
    # TEXT_ADA_001 = 500
    TEXT_EMBEDDING_ADA_002 = 8191

class ModelHardTokenLimit(ExtendedEnum):
    GPT_3_5_TURBO = 4000
    GPT_3_5_TURBO_0301 = 4000

    # TEXT_DAVINCI_003 = 4000
    # TEXT_CURIE_001 = 2048
    # TEXT_BABBAGE_001 = 2048
    # TEXT_ADA_001 = 2048
    TEXT_EMBEDDING_ADA_002 = 8191

class ModelTokenCost(ExtendedEnum):
    GPT_3_5_TURBO = 0.002
    GPT_3_5_TURBO_0301 = 0.002

    # TEXT_DAVINCI_003 = 0.02
    # TEXT_CURIE_001 = 0.002
    # TEXT_BABBAGE_001 = 0.0005
    # TEXT_ADA_001 = 0.0004

    TEXT_EMBEDDING_ADA_002 = 0.0004


class ModelTokenCostUnit(ExtendedEnum):
    THOUSAND = 1000


class ModelDescription(ExtendedEnum):
    GPT_3_5_TURBO = f"GPT 3.5 Turbo: {str(ModelHardTokenLimit.GPT_3_5_TURBO.value)} token Limit. ${str(ModelTokenCost.GPT_3_5_TURBO.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.     OpenAI's most powerful model. Longer outputs, more coherent, and more creative."
    GPT_3_5_TURBO_0301 = f"GPT 3.5 Turbo March 1st: {str(ModelHardTokenLimit.GPT_3_5_TURBO_0301.value)} token Limit. ${str(ModelTokenCost.GPT_3_5_TURBO_0301.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.     OpenAI's most powerful model. Longer outputs, more coherent, and more creative."

    # TEXT_DAVINCI_003 = f"Text Davinci 003: {str(ModelHardTokenLimit.TEXT_DAVINCI_003.value)} token Limit. ${str(ModelTokenCost.TEXT_DAVINCI_003.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.     OpenAI's most powerful model. Longer outputs, more coherent, and more creative."
    # TEXT_CURIE_001 = f"Text Curie 001:   {str(ModelHardTokenLimit.TEXT_CURIE_001.value)} token Limit. ${str(ModelTokenCost.TEXT_CURIE_001.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.    Very capable, but faster and lower cost than Davinci."
    # TEXT_BABBAGE_001 = f"Text Babbage 001: {str(ModelHardTokenLimit.TEXT_BABBAGE_001.value)} token Limit. ${str(ModelTokenCost.TEXT_BABBAGE_001.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.   Faster yet. Lower cost, but less compelling responses."
    # TEXT_ADA_001 = f"Text Ada 001:     {str(ModelHardTokenLimit.TEXT_ADA_001.value)} token Limit. ${str(ModelTokenCost.TEXT_ADA_001.value)} per {str(ModelTokenCostUnit.THOUSAND.value)} tokens.   OpenAI's fasted and cheap model. Not recommended for generating final responses."

    TEXT_EMBEDDING_ADA_002 = f"Text Embedding Ada 002: {str(ModelHardTokenLimit.TEXT_EMBEDDING_ADA_002)} token Limit. ${str(ModelTokenCost.TEXT_EMBEDDING_ADA_002)} per {str(ModelTokenCostUnit.THOUSAND)} tokens.   OpenAI's best advertised embedding model. Fast and cheap! Recommended for generating deep and shallow indexes"


## Service Models (By Type)


class ModelTextCompletionOpenAI(ExtendedEnum):
    GPT_3_5_TURBO = ModelID.GPT_3_5_TURBO.value
    GPT_3_5_TURBO_0301 = ModelID.GPT_3_5_TURBO_0301.value

    # TEXT_DAVINCI_003 = ModelID.TEXT_DAVINCI_003.value
    # TEXT_CURIE_001 = ModelID.TEXT_CURIE_001.value
    # TEXT_BABBAGE_001 = ModelID.TEXT_BABBAGE_001.value
    # TEXT_ADA_001 = ModelID.TEXT_ADA_001.value

class ModelTextEmbeddingOpenAI(ExtendedEnum):
    TEXT_EMBEDDING_ADA_002 = ModelID.TEXT_EMBEDDING_ADA_002.value


## Service Models (ALL)


class ModelOpenAI(ExtendedEnum):
    GPT_3_5_TURBO = ModelID.GPT_3_5_TURBO.value
    GPT_3_5_TURBO_0301 = ModelID.GPT_3_5_TURBO_0301.value

    # TEXT_DAVINCI_003 = ModelID.TEXT_DAVINCI_003.value
    # TEXT_CURIE_001 = ModelID.TEXT_CURIE_001.value
    # TEXT_BABBAGE_001 = ModelID.TEXT_BABBAGE_001.value
    # TEXT_ADA_001 = ModelID.TEXT_ADA_001.value

    TEXT_EMBEDDING_ADA_002 = ModelID.TEXT_EMBEDDING_ADA_002.value


MODEL_STATIC: dict = {
    ModelID.GPT_3_5_TURBO.value: {
        ModelParameterKey.ID.value: ModelID.GPT_3_5_TURBO.value,
        ModelParameterKey.SERVICE.value: ModelService.GPT_3_5_TURBO.value,
        ModelParameterKey.NAME.value: ModelName.GPT_3_5_TURBO.value,
        ModelParameterKey.MODEL_TYPE.value: ModelType.TEXT_COMPLETION.value,
        ModelParameterKey.SOFT_TOKEN_LIMIT.value: ModelSoftTokenLimit.GPT_3_5_TURBO.value,
        ModelParameterKey.HARD_TOKEN_LIMIT.value: ModelHardTokenLimit.GPT_3_5_TURBO.value,
        ModelParameterKey.TOKEN_COST.value: ModelTokenCost.GPT_3_5_TURBO.value,
        ModelParameterKey.TOKEN_COST_UNIT.value: ModelTokenCostUnit.THOUSAND.value,
    },
    ModelID.GPT_3_5_TURBO_0301.value: {
        ModelParameterKey.ID.value: ModelID.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.NAME.value: ModelName.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.SERVICE.value: ModelService.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.MODEL_TYPE.value: ModelType.TEXT_COMPLETION.value,
        ModelParameterKey.SOFT_TOKEN_LIMIT.value: ModelSoftTokenLimit.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.HARD_TOKEN_LIMIT.value: ModelHardTokenLimit.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.TOKEN_COST.value: ModelTokenCost.GPT_3_5_TURBO_0301.value,
        ModelParameterKey.TOKEN_COST_UNIT.value: ModelTokenCostUnit.THOUSAND.value,
    },
    ModelID.TEXT_EMBEDDING_ADA_002.value: {
        ModelParameterKey.ID.value: ModelID.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.NAME.value: ModelName.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.SERVICE.value: ModelService.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.MODEL_TYPE.value: ModelType.TEXT_EMBEDDING.value,
        ModelParameterKey.SOFT_TOKEN_LIMIT.value: ModelSoftTokenLimit.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.HARD_TOKEN_LIMIT.value: ModelHardTokenLimit.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.TOKEN_COST.value: ModelTokenCost.TEXT_EMBEDDING_ADA_002.value,
        ModelParameterKey.TOKEN_COST_UNIT.value: ModelTokenCostUnit.THOUSAND.value,
    }
}


ModelUnion = Union[
    ModelID,
    ModelParameterKey,
    ModelName,
    ModelHardTokenLimit,
    ModelDescription,
]


def get_model_static(static: Type[ModelUnion], key: ModelUnion) -> ModelUnion:
    return static.__members__[key.name]



import click

# from mindflow.cli.new_click_cli.commands.config import config
from mindflow.cli.new_click_cli.commands.ask import ask
from mindflow.cli.new_click_cli.commands.commit import commit
from mindflow.cli.new_click_cli.commands.delete import delete
from mindflow.cli.new_click_cli.commands.diff import diff
from mindflow.cli.new_click_cli.commands.index import index
from mindflow.cli.new_click_cli.commands.inspect import inspect
from mindflow.cli.new_click_cli.commands.login import login
from mindflow.cli.new_click_cli.commands.pr import pr
from mindflow.cli.new_click_cli.commands.query import query

@click.group()
def mindflow_cli():
    pass

@mindflow_cli.command()
def version():
    """Print the version of mindflow."""
    from mindflow import __version__
    click.echo(__version__)

mindflow_cli.add_command(ask)
mindflow_cli.add_command(commit)
mindflow_cli.add_command(delete)
mindflow_cli.add_command(diff)
mindflow_cli.add_command(index)
mindflow_cli.add_command(inspect)
mindflow_cli.add_command(login)
mindflow_cli.add_command(pr)
mindflow_cli.add_command(query)

if __name__ == '__main__':
    mindflow_cli()


from enum import Enum
from typing import List


class Command(Enum):
    """
    Arguments for MindFlow
    """

    ASK = "ask"
    COMMIT = "commit"
    CONFIG = "config"
    DELETE = "delete"
    DIFF = "diff"
    INDEX = "index"
    INSPECT = "inspect"
    QUERY = "query"
    REFRESH = "refresh"


class Arguments:
    """
    Arguments for the command line client
    """

    document_paths: List[str]
    force: bool
    index: bool
    diff_args: List[str]
    commit_args: List[str]
    query: str
    skip_clipboard: bool

    def __init__(self, params: dict):
        self.document_paths = params.get("document_paths", [])
        self.force = params.get("force", False)
        self.index = params.get("index", False)
        self.diff_args = params.get("diff_args", [])
        self.commit_args = params.get("commit_args", [])
        self.query = params.get("query", "")
        self.skip_clipboard = params.get("skip_clipboard", False)



"""
`inspect` command
"""
from typing import List

import click

from mindflow.core.inspect import run_inspect

@click.command(help="Inspect your MindFlow index")
@click.argument("document_paths", type=str, nargs=-1)
def inspect(document_paths: List[str]):
    """
    This function is used to inspect your MindFlow index.
    """
    run_inspect(document_paths)



from simple_term_menu import TerminalMenu
from typing import List


def menu(options: List[str], prompt: str):
    print(prompt)
    response = options[TerminalMenu(options).show()]
    print("\033[A\033[2K", end="")
    return response



"""
`generate` command
"""
from asyncio import Future
from copy import deepcopy

from typing import List, Optional
from concurrent.futures import ThreadPoolExecutor
import numpy as np

from alive_progress import alive_bar
from mindflow.db.controller import DATABASE_CONTROLLER
from mindflow.db.objects.model import Model

from mindflow.resolving.resolve import return_if_indexable, resolve_all


from mindflow.db.objects.document import Document, DocumentReference
from mindflow.settings import Settings
from mindflow.utils.document.read import read_document

from mindflow.utils.prompt_builders import build_context_prompt

from mindflow.utils.prompts import INDEX_PROMPT_PREFIX

def run_index(document_paths: List[str], refresh: bool, force: bool) -> None:
    """
    This function is used to generate an index and/or embeddings for files
    """

    import os
    document_paths = [os.path.abspath(path) for path in document_paths]

    settings = Settings()
    completion_model: Model = settings.mindflow_models.index.model
    
    indexable_document_references: List[DocumentReference] = return_if_indexable(resolve_all(document_paths), refresh, force)
    if not indexable_document_references:
        print("No documents to index")
        return

    total_size = sum(
        [
            document_reference.size
            for document_reference in indexable_document_references
        ]
    )
    print(f"Total content size: MB {total_size / 1024 / 1024:.2f}")
    # build search trees in parallel
    with alive_bar(
        len(indexable_document_references), bar="blocks", spinner="twirls"
    ) as progress_bar:

        with ThreadPoolExecutor(max_workers=50) as executor:
            search_tree_futures: List[Future[dict]] = [
                executor.submit(
                    create_text_search_tree,
                    completion_model,
                    read_document(
                        document_reference.path, document_reference.document_type
                    ),
                )
                for document_reference in indexable_document_references
            ]

            for search_tree_future, document_reference in zip(
                search_tree_futures, indexable_document_references
            ):
                document: Document = document_reference.to_document()
                document.search_tree = search_tree_future.result()
                document.save()
                del document_reference, search_tree_future
                progress_bar()
    
    DATABASE_CONTROLLER.databases.json.save_file()

class Node:
    """
    Simple node class for search tree.
    """

    start: int
    end: int
    summary: str
    embedding: np.ndarray
    leaves: List["Node"]

    def __init__(self, completion_model: Model, start: int, end: int, text: Optional[str] = None):
        self.start = start
        self.end = end
        if text:
            self.summary = completion_model(build_context_prompt(INDEX_PROMPT_PREFIX, text))

    def set_leaves(self, leaves: List["Node"]) -> None:
        self.leaves = leaves

    def to_dict(self) -> dict:
        return {
            "start": self.start,
            "end": self.end,
            "summary": self.summary if hasattr(self, "summary") else None,
            "embedding": self.embedding if hasattr(self, "embedding") else None,
            "leaves": [leaf.to_dict() for leaf in self.leaves]
            if hasattr(self, "leaves")
            else None,
        }

    def iterative_to_dict(self) -> dict:
        # Can't figure out what is wrong with this yet. Leaves coming up null.
        stack = [(self, None)]
        while stack:
            node, parent_leaves = stack.pop()
            node_dict: dict = {
                "start": node.start,
                "end": node.end,
                "summary": node.summary if hasattr(node, "summary") else None,
                "embedding": node.embedding if hasattr(node, "embedding") else None,
                "leaves": [],
            }
            if parent_leaves:
                parent_leaves.append(node_dict)
            if hasattr(node, "leaves"):
                for leaf in node.leaves:
                    leaves = node_dict["leaves"]
                    stack.append((leaf, leaves))
        return node_dict


def count_tokens(text: str) -> int:
    """
    Counts/estimates the number of tokens this text will consume by GPT.
    """
    # tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
    # count = len(tokenizer(text)['input_ids'])
    return len(text) // 4  # Token Estimation for speed


# This function is used to split a string into chunks of a specified token limit using binary search
def binary_split_raw_text_to_nodes(completion_model: Model, text: str) -> List[Node]:
    """
    Splits text into smaller chunks to not exceed the token limit.
    """
    nodes = []
    stack = [(0, len(text))]
    while stack:
        start, end = stack.pop()
        if count_tokens(text[start:end]) < completion_model.soft_token_limit:
            nodes.append(Node(completion_model, start, end, text[start:end]))
        else:
            mid = ((end - start) // 2) + start
            stack.append((start, mid))
            stack.append((mid, end))
    return nodes


def binary_split_nodes_to_chunks(completion_model: Model, nodes: List[Node]) -> List[List[Node]]:
    """
    Split nodes into smaller chunks to not exceed the token limit.
    """
    chunks = []
    stack = [(nodes, 0, len(nodes))]
    while stack:
        nodes, start, end = stack.pop()
        if (
            sum(count_tokens(node.summary) for node in nodes[start:end])
            < completion_model.soft_token_limit
        ):
            chunks.append(nodes[start:end])
        else:
            mid = (start + end) // 2
            stack.append((nodes, deepcopy(start), mid))
            stack.append((nodes, mid, deepcopy(end)))
    return chunks


def create_nodes(completion_model: Model, leaf_nodes: List[Node]) -> Node:
    """
    This function is used to iteratively create a nodes of our search tree
    """
    stack = [(leaf_nodes, 0, len(leaf_nodes))]
    while stack:
        leaf_nodes, start, end = stack.pop()
        if (
            sum(count_tokens(leaf_node.summary) for leaf_node in leaf_nodes[start:end])
            > completion_model.soft_token_limit
        ):
            node_chunks: List[List[Node]] = binary_split_nodes_to_chunks(
                completion_model,
                leaf_nodes[start:end]
            )
            for node_chunk in node_chunks:
                stack.append((node_chunk, 0, len(node_chunk)))
        else:
            parent_node_summaries = "\n".join(
                [node.summary for node in leaf_nodes[start:end]]
            )
            node = Node(
                completion_model,
                leaf_nodes[start].start,
                leaf_nodes[end - 1].end,
                parent_node_summaries,
            )
            node.set_leaves(leaf_nodes[start:end])
            return node
    return Node(completion_model, 0, 0)


def create_text_search_tree(completion_model: Model, text: str) -> dict:
    """
    This function is used to create a tree of responses from the OpenAI API
    """
    if count_tokens(text) < completion_model.soft_token_limit:
        return Node(completion_model, 0, len(text), text).to_dict()

    return create_nodes(completion_model, binary_split_raw_text_to_nodes(completion_model, text)).to_dict()



from typing import Dict, Optional

from mindflow.db.db.database import Database, Store
from mindflow.db.db.json import JsonDatabase
from mindflow.db.db.neo4j import Neo4jDatabase
from mindflow.db.db.static import Static


class Databases:
    def __init__(self):
        self.databases = {}

    @property
    def static(self) -> Static:
        if Store.STATIC not in self.databases:
            self.databases[Store.STATIC] = Static()
        return self.databases[Store.STATIC]
    
    @property
    def json(self) -> JsonDatabase:
        if Store.JSON not in self.databases:
            self.databases[Store.JSON] = JsonDatabase()
        return self.databases[Store.JSON]
    
    @property
    def neo4j(self) -> Neo4jDatabase:
        if Store.NEO4J not in self.databases:
            self.databases[Store.NEO4J] = Neo4jDatabase()
        return self.databases[Store.NEO4J]


class DatabaseController:
    def __init__(self, db_config: Optional[Dict] = None):
        self.databases = Databases()

DATABASE_CONTROLLER = DatabaseController()


"""
`delete` command
"""

from typing import List
from mindflow.core.delete import run_delete

import click

@click.command(help="Delete your MindFlow index")
@click.argument("document_paths", type=str, nargs=-1)
def delete(document_paths: List[str]):
    """
    This function is used to delete your MindFlow index.
    """
    run_delete(document_paths)



import getpass
import sys

from mindflow.cli.config.menu import menu
from mindflow.db.objects.mindflow_model import MindFlowModelConfig
from mindflow.db.objects.model import ModelConfig
from mindflow.db.objects.service import ServiceConfig
from mindflow.db.objects.static_definition.model import (
    ModelConfigParameterKey,
    ModelHardTokenLimit,
    get_model_static,
)
from mindflow.db.objects.static_definition.mind_flow_model import (
    MindFlowModelID,
)

from mindflow.db.objects.static_definition.service import (
    ServiceConfigParameterKey,
    ServiceID,
)
from mindflow.cli.config.questions.common.enums import (
    MainOption,
    YesNo,
)
from mindflow.cli.config.questions.common.questions import (
    ask_another_config,
)
from mindflow.cli.config.questions.service.questions import (
    ask_service,
    ask_service_config,
)
from mindflow.cli.config.questions.model.questions import (
    ask_mind_flow_model_type,
    ask_model_by_service,
    ask_model_config,
    ask_model_text_completion_openai,
    ask_model_text_embedding_openai,
)
from mindflow.state import STATE


def set_configuration():
    option_choice = menu(
        MainOption.values(), "Which options would you like to configure?"
    )
    match MainOption(option_choice):
        # Auth Config
        case MainOption.AUTHORIZE:
            config_service()
        # Model Config
        case MainOption.MIND_FLOW_MODEL:
            config_mindflow_model()
        case MainOption.MODEL:
            config_model()

    match YesNo(ask_another_config()):
        case YesNo.YES:
            set_configuration()


def config_service():
    service_key = ask_service()
    service_config_param = ask_service_config()

    match service_config_param:
        case ServiceConfigParameterKey.API_KEY:
            value = getpass.getpass("API Key: ")
        case ServiceConfigParameterKey.API_SECRET:
            value = getpass.getpass("API Secret: ")
        case _:
            print("Unrecognized model config parameter")
            sys.exit(1)
    
    service_config = ServiceConfig.load(f"{service_key.value}_config")
    if not service_config:
        service_config = ServiceConfig(id=f"{service_key.value}_config")

    setattr(service_config, service_config_param.value, value)
    service_config.save()

def config_model():
    service_key = ask_service()
    model_key = ask_model_by_service(service_key)
    model_config_param = ask_model_config()

    match model_config_param:
        case ModelConfigParameterKey.SOFT_TOKEN_LIMIT:
            model_hard_token_limit = get_model_static(
                ModelHardTokenLimit, model_key
            ).value
            while True:
                value = input("Soft Token Limit: ")
                if not value.isdigit():
                    print("Soft token limit must be a number")
                    continue
                value = int(value)
                if value > model_hard_token_limit:
                    print("Soft token limit cannot be greater than hard token limit")
                else:
                    break
        case _:
            print("Unrecognized model config parameter")
            sys.exit(1)
    
    model_configs = ModelConfig.load(f"{model_key.value}_config")
    if not model_configs:
        model_configs = ModelConfig(id=f"{model_key.value}_config")
    
    setattr(model_configs, model_config_param.value, value)
    model_configs.save()


def config_mindflow_model():
    mindflow_model_key = ask_mind_flow_model_type()
    match mindflow_model_key:
        case MindFlowModelID.INDEX:
            match ask_service():
                case ServiceID.OPENAI:
                    value = ask_model_text_completion_openai().value
        case MindFlowModelID.QUERY:
            match ask_service():
                case ServiceID.OPENAI:
                    value = ask_model_text_completion_openai().value
        case MindFlowModelID.EMBEDDING:
            match ask_service():
                case ServiceID.OPENAI:
                    value = ask_model_text_embedding_openai().value
    

    mindflow_model_config = MindFlowModelConfig.load(f"{mindflow_model_key.value}_config")
    if not mindflow_model_config:
        mindflow_model_config = MindFlowModelConfig(id=f"{mindflow_model_key.value}_config")

    setattr(mindflow_model_config, "model", value)
    mindflow_model_config.save()



"""
`diff` command
"""

from typing import Tuple
import click
from mindflow.core.diff import run_diff

@click.command(context_settings=dict(
    ignore_unknown_options=True,
), help="Wrapper around git diff that summarizes the output. Treat this command exactly like git diff, it supports all arguments that git diff provides.")
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def diff(args: Tuple[str]) -> str:
    print(run_diff(args))




"""
Arguments setters for CLI parser.
"""

import argparse
import sys

from mindflow.input import Command


# Command Parsers
def ask_args() -> argparse.Namespace:
    """
    Ask command parser.
    """
    parser = argparse.ArgumentParser(
        description="Ask a question.",
    )
    add_prompt_arg(parser)
    add_skip_clipboard_arg(parser)
    return parser.parse_args(sys.argv[2:])

def commit_args() -> argparse.Namespace:
    """
    Commit command parser.
    """
    parser = argparse.ArgumentParser(
        description="Generate commit message and commit.",
    )
    add_commit_args_arg(parser)
    add_skip_clipboard_arg(parser)
    return parser.parse_args(sys.argv[2:])


def config_args() -> argparse.Namespace:
    """
    Config command parser.
    """
    parser = argparse.ArgumentParser(
        description="Configure Mindflow.",
    )
    return parser.parse_args(sys.argv[2:])


def delete_args() -> argparse.Namespace:
    """
    Delete command parser.
    """
    parser = argparse.ArgumentParser(
        description="Delete a document.",
    )
    add_document_paths_arg(parser)
    return parser.parse_args(sys.argv[2:])


def diff_args() -> argparse.Namespace:
    """
    Diff command parser.
    """
    parser = argparse.ArgumentParser(
        description="Summarize your git diff.",
    )
    add_diff_args_arg(parser)
    add_skip_clipboard_arg(parser)
    return parser.parse_args(sys.argv[2:])


def index_args() -> argparse.Namespace:
    """
    Generate command parser.
    """
    parser = argparse.ArgumentParser(
        description="Generate an index.",
    )
    add_document_paths_arg(parser)
    return parser.parse_args(sys.argv[2:])


def inspect_args() -> argparse.Namespace:
    """
    Inspect command parser.
    """
    parser = argparse.ArgumentParser(
        description="Inspect your index.",
    )
    add_document_paths_arg(parser)
    return parser.parse_args(sys.argv[2:])


def query_args() -> argparse.Namespace:
    """
    Query command parser.
    """
    parser = argparse.ArgumentParser(
        description="Query your index.",
    )
    add_document_paths_arg(parser)
    add_index_arg(parser)
    add_query_arg(parser)
    add_skip_clipboard_arg(parser)
    return parser.parse_args(sys.argv[2:])


def refresh_args() -> argparse.Namespace:
    """
    Refresh command parser.
    """
    parser = argparse.ArgumentParser(
        description="Refresh your index.",
    )
    add_document_paths_arg(parser)
    add_force_arg(parser)
    return parser.parse_args(sys.argv[2:])


def add_index_arg(parser: argparse.ArgumentParser):
    """
    Add arguments for the generate command.
    """
    parser.add_argument(
        "-i",
        "--index",
        action="store_true",
        help="Generate an index for documents.",
    )


def add_document_paths_arg(parser: argparse.ArgumentParser):
    """
    Add arguments for handling document paths.
    """
    parser.add_argument(
        "document_paths",
        nargs="+",
        help="A list of document paths to summarize (file path, API, web address).",
    )

def add_prompt_arg(parser: argparse.ArgumentParser):
    """
    Add prompt argument..
    """
    parser.add_argument(
        "prompt", type=str, help="Prompt you'd like to sent off to LLM."
    )

def add_query_arg(parser: argparse.ArgumentParser):
    """
    Add arguments for querying documents.
    """
    parser.add_argument(
        "query", type=str, help="The query you want to make on some data."
    )


def add_commit_args_arg(parser: argparse.ArgumentParser):
    """
    Add arguments for the commit command.
    """
    pass


def add_diff_args_arg(parser: argparse.ArgumentParser):
    """
    Add arguments for the diff command.
    """
    parser.add_argument('-r', '--revision', metavar='REV', help='Specify the revision to diff against')
    parser.add_argument('-s', '--summary', action='store_true', help='Show summary of changes')
    parser.add_argument('file', nargs='*', help='File or directory to diff')


def add_skip_clipboard_arg(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-s",
        "--skip-clipboard",
        action="store_true",
        help="Do not copy to clipboard (testing).",
    )


def add_force_arg(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Apply THE FORCE to your command.",
    )


def get_parsed_cli_args(command: str) -> argparse.Namespace:
    match command:
        case Command.ASK.value:
            return ask_args()
        case Command.COMMIT.value:
            return commit_args()
        case Command.CONFIG.value:
            return config_args()
        case Command.DELETE.value:
            return delete_args()
        case Command.DIFF.value:
            return diff_args()
        case Command.INDEX.value:
            return index_args()
        case Command.INSPECT.value:
            return inspect_args()
        case Command.QUERY.value:
            return query_args()
        case Command.REFRESH.value:
            return refresh_args()
        case _:
            raise ValueError(f"Invalid command: {command}")



from typing import Union
from mindflow.db.objects.static_definition.mind_flow_model import MINDFLOW_MODEL_STATIC

from mindflow.db.objects.static_definition.service import SERVICE_STATIC
from mindflow.db.objects.static_definition.model import MODEL_STATIC

from mindflow.db.db.database import Database, Collection


class Static(Database):
    @staticmethod
    def load(collection: str, object_key: str):
        match collection:
            case Collection.SERVICE.value:
                return SERVICE_STATIC[object_key]
            case Collection.MODEL.value:
                return MODEL_STATIC[object_key]
            case Collection.MIND_FLOW_MODEL.value:
                return MINDFLOW_MODEL_STATIC[object_key]
            case _:
                raise ValueError(f"Unknown object collection: {collection}")



"""
`ask` command
"""

import click

from mindflow.core.ask import run_ask

@click.command(help="")
@click.argument("prompt", type=str)
def ask(prompt: str) -> str:
    """
    This function is used to generate a prompt and then use it as a prompt for GPT bot.
    """
    print(run_ask(prompt))



from setuptools import setup, find_packages

# get version from mindflow/__init__.py
with open("mindflow/__init__.py", "r") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip('"')
            break

with open("requirements.txt", "r") as f:
    install_requires = f.read().strip().split("\n")

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mindflow",
    python_requires='>=3.10',
    version=version,
    py_modules=["mindflow"],
    # entry_points={"console_scripts": ["mf = mindflow.main:main"]},
    entry_points={"console_scripts": ["mf = mindflow.cli.new_click_cli.cli_main:mindflow_cli"]},
    packages=find_packages(),
    install_requires=install_requires,
    description="AI-powered search engine for your code!",
    long_description=long_description,
    long_description_content_type="text/markdown",
)




from typing import List, Optional
from mindflow.utils.enum import ExtendedEnum

class Store(ExtendedEnum):
    STATIC = "static"
    JSON = "json"
    NEO4J = "neo4j"


class Collection(ExtendedEnum):
    SERVICE = "service"
    MODEL = "model"
    MIND_FLOW_MODEL = "mind_flow_model"
    CONFIGURATIONS = "configurations"
    DOCUMENT = "document"

class Database:
    def load(self, collection: str, object_id: str) -> Optional[dict]:
        raise NotImplementedError

    def load_bulk(
        self, collection: str, object_ids: List[str]
    ) -> Optional[List[dict]]:
        raise NotImplementedError

    def delete(self, collection: str, object_id: str):
        raise NotImplementedError

    def delete_bulk(self, collection: str, object_ids: List[str]):
        raise NotImplementedError

    def save(self, collection: str, value: dict):
        raise NotImplementedError



from typing import Union
from mindflow.utils.enum import ExtendedEnum


class ModelType(ExtendedEnum):
    TEXT_COMPLETION = "text_completion"
    TEXT_EMBEDDING = "text_embedding"


class ModelTypeName(ExtendedEnum):
    TEXT_COMPLETION = "Text Completion"
    TEXT_EMBEDDING = "Text Embedding"


## Service Model Types
class ModelTypeOpenAI(ExtendedEnum):
    TEXT_COMPLETION = ModelType.TEXT_COMPLETION.value
    TEXT_EMBEDDING = ModelType.TEXT_EMBEDDING.value


ModelTypeUnion = Union[
    ModelType,
    ModelTypeName,
    ModelTypeOpenAI,
]


def get_model_type_static(
    static: ModelTypeUnion, key: ModelTypeUnion
) -> ModelTypeUnion:
    return static.__members__[key.name]



"""
Module for recursively extracting all files from a file path.
"""

import logging
import os

from typing import List, Union

from mindflow.utils.files.git import is_within_git_repo, get_git_files


def extract_files(path: Union[str, os.PathLike]) -> List[str]:
    """
    Extract all files from a directory.
    """
    file_paths = []
    if os.path.isfile(path):
        return [os.fspath(path)]
    if is_within_git_repo(path):
        return get_git_files(path)
    try:
        with os.scandir(path) as paths:
            for scanned_path in paths:
                if scanned_path.is_file():
                    file_paths.append(os.path.abspath(scanned_path.path))
                elif os.path.isdir(path):
                    file_paths.extend(extract_files(scanned_path.path))
        return file_paths
    except IsADirectoryError as error:
        logging.debug("Could not read directory/file: %s", path)
        logging.debug(error)
        return file_paths



"""
`query` command
"""

from typing import List
import click

from mindflow.core.query import run_query

@click.command()
@click.argument("document_paths", type=str, nargs=-1, required=True)
@click.argument("query", type=str, required=True)
# def query(document_paths: List[str], query: str, completion_model: bool, embedding_model: Model):
def query(document_paths: List[str], query: str):
    print(run_query(document_paths, query))


from mindflow.settings import Settings
from mindflow.utils.response import handle_response_text


def run_ask(prompt: str) -> str:
    """
    This function is used to generate a prompt and then use it as a prompt for GPT bot.
    """
    settings = Settings()
    # Prompt GPT through Mindflow API or locally
    response: str = settings.mindflow_models.query.model([{"role": "system", "content": "You are a helpful virtual assistant responding to a users query using your general knowledge and the text provided below."}, {"role": "user", "content": prompt}])
    return response



"""
Command Line Client for Mindflow
"""

import argparse
import sys

from mindflow.db.controller import DATABASE_CONTROLLER
from mindflow.state import STATE
from mindflow.input import Arguments, Command

from mindflow.cli.parser import get_parsed_cli_args
from mindflow.settings import Settings

from mindflow.core.ask import AskArgs, ask
from mindflow.cli.new_click_cli.commands.commit import commit
from mindflow.cli.new_click_cli.commands.config import config
from mindflow.cli.new_click_cli.commands.delete import DeleteArgs, delete
from mindflow.cli.new_click_cli.commands.diff import DiffArgs, diff
from mindflow.cli.new_click_cli.commands.index import IndexArgs, index
from mindflow.cli.new_click_cli.commands.inspect import InspectArgs, inspect
from mindflow.core.query import QueryArgs, query
from mindflow.utils.response import handle_response_text

MF_DESCRIPTION = """

Welcome to Mindflow. A command line tool for intelligent development and collaboration.

"""

MF_USAGE = """

mf <command> [<args>]

ask        `mf ask <PROMPT>`                              Ask a question and get a prompt.
commit     `mf commit [<git commit args>]`                Generate commit message and commit.
config     `mf config`                                    Configure Mindflow.
diff       `mf diff [<git diff args>]`                    Runs a `git diff` and summarizes the changes.
index      `mf index [<document paths>]`                  Generate index of files/folders.
query      `mf query <YOUR QUERY> [<document paths>]`     Query your files/folders.
delete     `mf delete <document paths>`                   Delete a file/folder from Mindflow.
refresh    `mf refresh <document paths>`                  Refresh Mindflow index by regenerating already existing document indexes.
inspect    `mf inspect <document paths>`                  Inspect a file/folder in the Mindflow index.

"""


def cli():
    # Parse Arguments
    parser = set_parser()
    args = parser.parse_args(arg.upper() for arg in sys.argv[1:2])
    command = Command[args.command].value
    args = get_parsed_cli_args(command)

    # Configure State
    settings = Settings()

    # Execute Command
    match command:
        case Command.ASK.value:
            response: str = ask(args.prompt, settings.mindflow_models.query)
            handle_response_text(response, )
        case Command.COMMIT.value:
            commit()
        case Command.CONFIG.value:
            config()
        case Command.DELETE.value:
            delete(args.document_paths)
        case Command.DIFF.value:
            diff(settings.mindflow_models.index)
        case Command.INSPECT.value:
            inspect(args.document_paths)
        case Command.QUERY.value:
            if hasattr(args, "index") and args.index:
                index(args.document_paths, Command.INDEX, args.force, settings.mindflow_models.index)
            query(args.document_paths, args.query, settings.mindflow_models.query, settings.mindflow_models.embedding)
        case Command.REFRESH.value:
            index(args.document_paths, Command.INDEX, args.force, settings.mindflow_models.index)
        case Command.INDEX.value:
            index(args.document_paths, Command.REFRESH, args.force, settings.mindflow_models.index)

    print("Saving database...")
    DATABASE_CONTROLLER.databases.json.save_file()


def set_parser():
    parser = argparse.ArgumentParser(description=MF_DESCRIPTION, usage=MF_USAGE)
    parser.add_argument(
        "command",
        choices=Command.__members__,
        help="The command to execute",
    )
    return parser



from mindflow.cli.config.main import menu
from mindflow.cli.config.questions.common.enums import YesNo


def ask_another_config() -> YesNo:
    continue_choice = menu(YesNo.values(), "Would you like to configure another item?")
    return YesNo(continue_choice)
